#include <afx.h>
#include <memory.h>
#include <stdio.h>
#include "OBBitmap.h"



unsigned int nBitmapInfoSize = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 256;
unsigned int nFileHeaderSize = nBitmapInfoSize + sizeof(BITMAPFILEHEADER);



__declspec(dllexport) OBBitmap::OBBitmap(){
	memset(this, 0, sizeof(OBBitmap));

	BMI = (BITMAPINFO*) new unsigned char[sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 256];
	memset(BMI, 0, sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 256);
	BMI->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	BMI->bmiHeader.biPlanes = 1;		// Must always be 1 according to docs
	BMI->bmiHeader.biBitCount = 8;
}



__declspec(dllexport) OBBitmap::~OBBitmap(){
	delete Data;
	delete [] BMI;
}



void __declspec(dllexport) OBBitmap::ProcessData(unsigned char *RawData, unsigned char nDefaultColor){
	unsigned int x;
	unsigned int y;

	memset(Data, nDefaultColor, nWidth * nHeight);
	
	if(nBitPlanes){
		for(y = 0; y < nHeight; y += 8){
			for(x = 0; x < nWidth; x += 8){
				Put8x8Chunk(x, y, &RawData[((y * nWidth) >> 1) + (x << 2)]);
			}
		}
	}
	else if(!nBitPlanes){
		for(y = 0; y < nHeight; y++){
			for(x = 0; x < nWidth; x++){
				Data[(nHeight - 1 - y) * nWidth + x] = RawData[(y * nWidth) + x];
			}
		}
	}
}



void __declspec(dllexport) OBBitmap::Put8x8Chunk(int dx, int dy, unsigned char *RawData, unsigned char PaletteShift, unsigned char bFlip){
	unsigned char px1;
	unsigned char px2;

	if(nBitPlanes == 3 || nBitPlanes == 4){
		for(int y = 0; y < 8; y++){
			for(int i = 7; i > -1; i--){
				unsigned char mask = 1 << i;
				px1 = ((RawData[y*2] & mask) >> i) | (((RawData[y*2+1] & mask) >> i) << 1);
				if(nBitPlanes != 3){
					px1 |=  (((RawData[y*2+16] & mask) >> i) << 2) | (((RawData[y*2+17] & mask) >> i) << 3);
				}
				else{
					px1 |=  (((RawData[y+16] & mask) >> i) << 2);
				}
				if(px1 != 0){
					unsigned int nXCoord = dx + (7 - i);
					unsigned int nYCoord = (nHeight - 1 - (y + dy)) * nWidth;
					if (bFlip & 0x01){
						nXCoord = dx + i;
					}
					if(bFlip & 0x02){
						nYCoord = (nHeight - 1 - ((7 - y) + dy)) * nWidth;
					}
					Data[nYCoord + nXCoord] = px1 | (PaletteShift << 4);
				}
			}
		}
	}
	else if(nBitPlanes == 1){
		for(int y = 0; y < 8; y++){
			for(int x = 0; x < 4; x++){
				px1 = RawData[(y << 2) | x] & 0x0F;
				px2 = (RawData[(y << 2) | x]) >> 4;
				Data[(nHeight - 1 - (y + dy)) * nWidth + dx + (2 * x)] = px1 | (PaletteShift << 4);
				Data[(nHeight - 1 - (y + dy)) * nWidth + dx + (2 * x) + 1] = px2 | (PaletteShift << 4);
			}
		}
	}
}



void __declspec(dllexport) OBBitmap::WriteOutBitmap(char *sFileName){
    BITMAPFILEHEADER hdr;
    hdr.bfType = 0x4D42;
    hdr.bfSize = nFileHeaderSize + nWidth * nHeight;
    hdr.bfReserved1 = 0;
    hdr.bfReserved2 = 0;
    hdr.bfOffBits = nFileHeaderSize;

	CFile Fout(sFileName, CFile::modeCreate | CFile::modeWrite | CFile::shareDenyWrite);
	Fout.Write(&hdr, sizeof(BITMAPFILEHEADER));
	Fout.Write(BMI, sizeof(BITMAPINFO) + sizeof(RGBQUAD) * 256);
	Fout.Write(Data, nWidth * nHeight);
	Fout.Close();
}



void __declspec(dllexport) OBBitmap::SetDim(unsigned int width, unsigned int height){
	BMI->bmiHeader.biWidth = nWidth = width;
	BMI->bmiHeader.biHeight = nHeight = height;
	if(Data){
		delete Data;
	}
	Data = new unsigned char[nWidth * nHeight];
}



void __declspec(dllexport) OBBitmap::SetPalette(unsigned short *RawPalette, unsigned short nColors){
	int j;

	if(nColors > 256){
		return;
	}

	for(j = 0; j < nColors; j++){
		BMI->bmiColors[j].rgbRed = (RawPalette[j] & 0x001F) << 3;
		BMI->bmiColors[j].rgbGreen = (RawPalette[j] & 0x03E0) >> 2;
		BMI->bmiColors[j].rgbBlue = (RawPalette[j] & 0x7C00) >> 7;
	}

	if(nColors < 256){
		BMI->bmiColors[j].rgbRed = 255;
		BMI->bmiColors[j].rgbGreen = 0;
		BMI->bmiColors[j].rgbBlue = 255;
	}
}



void __declspec(dllexport) OBBitmap::SetPaletteGrey(unsigned short nColors){
	int j;

	for(j = 0; j < nColors && j < 256; j++){
		BMI->bmiColors[j].rgbRed = j * (255 / (nColors - 1));
		BMI->bmiColors[j].rgbGreen = j * (255 / (nColors - 1));
		BMI->bmiColors[j].rgbBlue = j * (255 / (nColors - 1));
	}
}



void __declspec(dllexport) OBBitmap::SetDefaultColor(unsigned char nDefaultColor){
	memset(Data, nDefaultColor, nWidth * nHeight);
}



void __declspec(dllexport) OBBitmap::ResetTransparency(unsigned char nTransIndex, unsigned char nDefaultColor){
	unsigned int i;

	for(i = 0; i < nWidth * nHeight; i++){
		if(Data[i] == nTransIndex){
			Data[i] = nDefaultColor;
		}
	}
}
