#include <afx.h>
#include <fstream>
#include "waveform.h"

__declspec(dllexport) Waveform::Waveform(){
	sChunkID = "RIFF";
	sWaveID = "WAVE";
	sFormatID = "fmt ";
	nFormatSize = 0x10;
	nFormatTag = 1;
	nChannels = 1;
	nSamplesPerSec = 8000;
	nAvgBytesPerSec = 16000;
	nBlockAlign = 2;
	nBitsPerSample = 0x10;
	sDataID = "data";
	nData = 0;
}

__declspec(dllexport) Waveform::~Waveform(){
	delete nData;
}

void __declspec(dllexport) Waveform::SetDataSize(unsigned int nSize){
	if(nData){
		delete nData;
	}
	nData = new short[nSize];
	nDataSize = nSize;
	nChunkSize = nSize + 0x24;
}
void __declspec(dllexport) Waveform::SetSampleRate(unsigned int nSampleRate){
	nSamplesPerSec = nSampleRate;
	nAvgBytesPerSec = nSampleRate << 1;
}

void __declspec(dllexport) Waveform::DecodeBRR(unsigned char *nRawData){
	int i;
	int j;
	unsigned short nLength = nRawData[0] | (nRawData[1] << 8);
	unsigned char bEndBit = 0;
	unsigned char nFilter;
	unsigned char nRange;
	unsigned char bLoopBit;
	int k = 0;

	for(i = 2; (unsigned) i < nLength && !bEndBit && (unsigned) i < nDataSize; i += 9){
		bEndBit = nRawData[i] & 0x01;
		bLoopBit = (nRawData[i] & 0x02) >> 1;
		nFilter = (nRawData[i] & 0x0C) >> 2;
		nRange = (nRawData[i] & 0xF0) >> 4;
		
		for(j = 1; j < 9; j++){
			nData[k] = (nRawData[i + j] >> 4) & 0x0F;  // Get the first nybble of the byte
			if(nData[k] > 7){             // The nybble is negative, so make the 
				nData[k] |= 0xFFF0;  // 16-bit value negative
			}
			nData[k] <<= nRange;    // Apply the RANGE value

			if(nFilter == 1){
				nData[k] += (short) ((double) nData[k - 1] * 15/16);
			}
			else if(nFilter == 2){
				nData[k] += (short) (((double) nData[k - 1] * 61/32) - ((double) nData[k - 2] * 15/16));
			}
			else if(nFilter == 3){
				nData[k] += (short) (((double) nData[k - 1] * 115/64) - ((double) nData[k - 2] * 13/16));
			}

			k++;                  // Advance our output pointer

			// Now do the same thing for the other nybble
			nData[k] = nRawData[i + j] & 0x0F; // Get the second nybble of the byte
			if(nData[k] > 7){             // The nybble is negative, so make the 
				nData[k] |= 0xFFF0;  // 16-bit value negative
			}
			nData[k] <<= nRange;    // Apply the RANGE value

			if(nFilter == 1){
				nData[k] += (short) ((double) nData[k - 1]/16*15);
			}
			else if(nFilter == 2){
				nData[k] += (short) (((double) nData[k - 1] * 61/32) - ((double) nData[k - 2] * 15/16));
			}
			else if(nFilter == 3){
				nData[k] += (short) (((double) nData[k - 1] * 115/64) - ((double) nData[k - 2] * 13/16));
			}

			k++;                  // Advance our output pointer
		}
	}
}



void __declspec(dllexport) Waveform::WriteOutWaveform(char *sFileName){
	ofstream Fout;
	Fout.open(sFileName, ios_base::trunc | ios_base::binary);

	Fout.write((char*) sChunkID.c_str(), 4);			   // chunk	id 'RIFF'
	Fout.write((char*) &nChunkSize, 4);			   // chunk	size
	Fout.write((char*) sWaveID.c_str(), 4);		   // wave chunk id	'WAVE'
	Fout.write((char*) sFormatID.c_str(), 4);		   // format chunk id 'fmt '
	Fout.write((char*) &nFormatSize, 4);		   // format chunk size
	Fout.write((char*) &nFormatTag, 2);		   // format tag currently pcm
	Fout.write((char*) &nChannels, 2);		   // number of channels 1=mono, 2=stereo
	Fout.write((char*) &nSamplesPerSec, 4);	   // sample rate in hz
	Fout.write((char*) &nAvgBytesPerSec, 4);	   // average bytes	per	second
	Fout.write((char*) &nBlockAlign, 2);		   // number of bytes per sample (1 = 8-bit mono,  2 = 8-bit stereo or 16-bit mono, 4 = 16-bit stereo)
	Fout.write((char*) &nBitsPerSample, 2);	   // number of	bits in	a sample
	Fout.write((char*) sDataID.c_str(), 4);		   // data chunk id	'data'
	Fout.write((char*) &nDataSize, 4);		   // length of	data chunk
	Fout.write((char*) nData, nDataSize);
	Fout.close();
}