// FF3Doc.cpp : implementation of the CFF3Doc class
//

#include <iostream>
#include <fstream>
#include <afxwin.h>
#include "FF3Doc.h"
#include "EPFF6Extract.h"
#include "OBBitmap.h"
#include "Waveform.h"
#include "mylib.h"

using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CFF3Doc construction/destruction

CFF3Doc::CFF3Doc(){
	worldData = 0;
	RomData = new byte[0x300200];
}

CFF3Doc::~CFF3Doc(){
	delete [] RomData;
}

/////////////////////////////////////////////////////////////////////////////
// CFF3Doc commands

void CFF3Doc::DeleteContents(){
	m_bValid = 0;
	
	if(worldData) { 
		delete [] worldData; 
		worldData = 0; 
	}

	for(int i = 0; i < dialogue.GetSize(); i++){
		delete dialogue[i];
	}
	dialogue.RemoveAll();

	CDocument::DeleteContents();
}



byte CFF3Doc::OnOpenDocument(char *lpszPathName){
	int i;
//	int j;
//	char filename[64];
	
	CFileException fe;
	Fin = GetFile(lpszPathName, CFile::modeRead|CFile::shareDenyWrite, &fe);
	if (!Fin){
		ReportSaveLoadException(lpszPathName, &fe, 0, AFX_IDP_FAILED_TO_OPEN_DOC);
		
		m_bValid = 0;
		return 0;
	}
	
	Fin->SeekToBegin();
	Fin->Read(&RomData[0x0200], 0x300000);

	DeleteContents();
	SetModifiedFlag();  // dirty during de-serialize
	
	// Check to see if the file is valid (ie, the American FF3 ROM and version 1.0).
	char szV10Title[22] = "FINAL FANTASY 3      ";
	
	if(memcmp(szV10Title, &RomData[0x0101C0], 21)){
		AfxMessageBox("The ROM you are attempting to load\nis not the American FF3 ROM.");
		m_bValid = 0;
		return 0;
	}
	
	if(RomData[0x0101DB]){
		AfxMessageBox("The ROM you are attempting to load\nis not version 1.0 of the American\nFF3 ROM.  Let me know if you see this, because\nI was not aware of other versions of\nFF3 being available.");
		m_bValid = 0;
		return 0;
	}
	
	byte RawData[65536];

	//Build greyscale palette
	ushort UnknownPalette[16];
	for(i = 0; i < 16; i++){
		UnknownPalette[i] = (i << 11) | (i << 6) | (i << 1);
	}

	WriteLocPropData();
	WriteLocStrings();

	//WriteOutEsperData(RomData);

/*	ofstream Fout;
	Fout.open("spell progression.txt");
	Fout << "Character\tLevel\tSkill\n";
	for(i = 0; i < 16; i++){
		Fout << "Terra\t" << (ushort) RomData[0x2CE5C1 + i] << "\t" << (ushort) RomData[0x2CE5C0 + i] << "\n";
		Fout << "Celes\t" << (ushort) RomData[0x2CE5E1 + i] << "\t" << (ushort) RomData[0x2CE5E0 + i] << "\n";
	}
	for(i = 0; i < 8; i++){
		Fout << "Sabin\t" << (ushort) RomData[0x26F690 + i] << "\t" << (i + 1) << "\n";
		Fout << "Cyan\t" << (ushort) RomData[0x26F698 + i] << "\t" << (i + 1) << "\n";
	}
	Fout.close();*/

/*		
	// Load the actual file.
	worldData = new WorldData[3];
	
	worldData[MAP_BALANCE].pDoc = this;
	worldData[MAP_BALANCE].size = CSize(256, 256);
	worldData[MAP_BALANCE].name = "World of Balance";
	worldData[MAP_BALANCE].mapPtrs.Add(0xEEB20F);
	worldData[MAP_BALANCE].miniMapPtrs.Add(0xEEB24B);
	worldData[MAP_BALANCE].tilePtrs.Add(0xEEB212);
	worldData[MAP_BALANCE].ReadMapData(0x2ED634);
	//	worldData[MAP_BALANCE].WriteOutMapData("mapdata\\wobmapdata.txt");
	worldData[MAP_BALANCE].ReadMiniMapData(0x2FE69B);
	worldData[MAP_BALANCE].ReadPalettes(0x12EE00);
	worldData[MAP_BALANCE].ReadTileData(0x2F134F);
	worldData[MAP_BALANCE].ReadTileProperties(0x2E9D14);
	//	worldData[MAP_BALANCE].WriteOutTiles("wob");
	worldData[MAP_BALANCE].ReadTriggers(0x1FBD00, 0x040200);
	//	worldData[MAP_BALANCE].ReadEnemyLayout(0x74742, 64);
	//	worldData[MAP_BALANCE].ReadEnemyOdds(0x74500, 64);
	
	worldData[MAP_RUIN].pDoc = this;
	worldData[MAP_RUIN].size = CSize(256, 256);
	worldData[MAP_RUIN].name = "World of Ruin";
	worldData[MAP_RUIN].mapPtrs.Add(0xEEB221);
	worldData[MAP_RUIN].mapPtrs.Add(0xEEB224);
	worldData[MAP_RUIN].miniMapPtrs.Add(0xEEB24E);
	worldData[MAP_RUIN].tilePtrs.Add(0xEEB21E);
	worldData[MAP_RUIN].ReadMapData(0x2F6C56);
	//	worldData[MAP_RUIN].WriteOutMapData("mapdata\\wormapdata.txt");
	worldData[MAP_RUIN].ReadMiniMapData(0x2FEAB3);
	WriteOutMiniMap();
	worldData[MAP_RUIN].ReadPalettes(0x12EF00);
	worldData[MAP_RUIN].ReadTileData(0x2F4C46);
	worldData[MAP_RUIN].ReadTileProperties(0x2E9F14);
	//	worldData[MAP_RUIN].WriteOutTiles("wor");
	worldData[MAP_RUIN].ReadTriggers(0x1FBD02, 0x040202);
	//	worldData[MAP_RUIN].ReadEnemyLayout(0x74742, 64);
	//	worldData[MAP_RUIN].ReadEnemyOdds(0x74500, 64);
*/	

	OBBitmap LocMap;
	LocMap.nBitPlanes = 4;
	uint nHeight;
	uint nWidth;
	uint nLocMap1Height;
	uint nLocMap1Width;
	uint nLocMap2Height;
	uint nLocMap2Width;
	uint nLocMapDataAddr;
	byte* LocMapData = new byte[0x4000];
	byte* LocMapData2 = new byte[0x4000];
	byte* nLocMapRawData = new byte[0x400000];
	byte nTile = 0;
	byte nTile2 = 0;

	OBBitmap LocMapTiles;
	LocMapTiles.nBitPlanes = 4;
	LocMapTiles.SetDim(256, 512);
	byte* nLocMapForm1 = new byte[0x800];
	byte* nLocMapForm2 = new byte[0x800];
	byte nLocMapForm1Index;
	byte nLocMapForm2Index;
	uint nLocMapForm1Addr;
	uint nLocMapForm2Addr;
	byte* nLocMapTiles = new byte[65536];
	byte nLocTileForm[8];
	int j;
	int k;
	byte nTileSet = 0;
	byte* nChunk;
	byte nPaletteShift = 0;
	byte nFlip = 0;
	byte nPaletteShift2 = 0;
	byte nFlip2 = 0;

	uint nAddr = 0;
	OBBitmap LocTiles;
	LocTiles.nBitPlanes = 4;
	LocTiles.SetPaletteGrey(16);
	LocTiles.SetDim(128, 608);
	char sFilename[256];

	OBBitmap nAnimatedTiles;
	nAnimatedTiles.SetDim(16, 512);
	uint nPalShift = 0;
	nAnimatedTiles.SetPaletteGrey(16);
	nAnimatedTiles.nBitPlanes = 4;

	for(uint nLocCtr = 0; nLocCtr < 415; nLocCtr++){
		nAddr = (*((uint*) &RomData[0x1FDC00 + (3 * (RomData[0x2D9107 + 33 * nLocCtr] & 0x7F))]) & 0xFFFFFF) + 0x1FDD00;
		memcpy(RawData, &RomData[nAddr], 0x2000);
		nAddr = (*((uint*) &RomData[0x1FDC00 + (3 * ((*((ushort*) &RomData[0x2D9107 + 33 * nLocCtr]) & 0x3F80) >> 7))]) & 0xFFFFFF) + 0x1FDD00;
		memcpy(&RawData[0x2000], &RomData[nAddr], 0x1000);
		nAddr = (*((uint*) &RomData[0x1FDC00 + (3 * ((*((ushort*) &RomData[0x2D9108 + 33 * nLocCtr]) & 0x1FC0) >> 6))]) & 0xFFFFFF) + 0x1FDD00;
		memcpy(&RawData[0x3000], &RomData[nAddr], 0x1000);
		nAddr = (*((uint*) &RomData[0x1FDC00 + (3 * ((*((ushort*) &RomData[0x2D9109 + 33 * nLocCtr]) & 0x0FE0) >> 5))]) & 0xFFFFFF) + 0x1FDD00;
		memcpy(&RawData[0x4000], &RomData[nAddr], 0x2000);
//		nAddr = 0x26CFC0 + (*((uint*) &RomData[0x26CFA0 + (((RomData[0x2D911B + 33 * nLocCtr] & 0xE0) >> 5) - 1) * 3]) & 0xFFFFFF);
//		uint nLen = DecompressData(&RomData[nAddr], &RawData[0x8800], 65536);

//		LocTiles.ProcessData(RawData, 0);
//		LocTiles.WriteOutBitmap("LocMapRawTiles.bmp");

		//nAnimatedTiles.SetPalette((ushort*) &RomData[0x2DC680 + (RomData[0x2D9119 + 660] << 8)], 128);
		//for(i = 0; i < 8; i++){
		//	//nPalShift = RomData[0x009400 + *((ushort*) &RomData[((RomData[0x2D911B + 33 * nLocCtr] & 0x1F) << 1) + 0x0093D5])];
		//	nAddr = 0x260200 + *((ushort*) &RomData[0x009401 + (i * 10) + *((ushort*) &RomData[((RomData[0x2D911B + 33 * nLocCtr] & 0x1F) << 1) + 0x0093D5])]);
		//	nAnimatedTiles.Put8x8Chunk(0, i << 6, &RomData[nAddr], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, i << 6, &RomData[nAddr + 32], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 8, &RomData[nAddr + 64], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 8, &RomData[nAddr + 96], nPalShift, 0);
		//	
		//	nAddr = 0x260200 + *((ushort*) &RomData[0x009403 + (i * 10) + *((ushort*) &RomData[((RomData[0x2D911B + 33 * nLocCtr] & 0x1F) << 1) + 0x0093D5])]);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 16, &RomData[nAddr], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 16, &RomData[nAddr + 32], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 24, &RomData[nAddr + 64], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 24, &RomData[nAddr + 96], nPalShift, 0);
		//	
		//	nAddr = 0x260200 + *((ushort*) &RomData[0x009405 + (i * 10) + *((ushort*) &RomData[((RomData[0x2D911B + 33 * nLocCtr] & 0x1F) << 1) + 0x0093D5])]);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 32, &RomData[nAddr], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 32, &RomData[nAddr + 32], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 40, &RomData[nAddr + 64], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 40, &RomData[nAddr + 96], nPalShift, 0);
		//	
		//	nAddr = 0x260200 + *((ushort*) &RomData[0x009407 + (i * 10) + *((ushort*) &RomData[((RomData[0x2D911B + 33 * nLocCtr] & 0x1F) << 1) + 0x0093D5])]);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 48, &RomData[nAddr], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 48, &RomData[nAddr + 32], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(0, (i << 6) + 56, &RomData[nAddr + 64], nPalShift, 0);
		//	nAnimatedTiles.Put8x8Chunk(8, (i << 6) + 56, &RomData[nAddr + 96], nPalShift, 0);
		//}
		//nAnimatedTiles.WriteOutBitmap("NarsheAnimatedTiles.bmp");

		nLocMapForm1Index = ((*((ushort*) &RomData[0x2D910B + 33 * nLocCtr]) & 0x01FC) >> 2);
		nLocMapForm2Index = ((RomData[0x2D910C + 33 * nLocCtr] & 0xFE) >> 1);

		nLocMapForm1Addr = 0x1E0200 + (*((uint*) &RomData[0x1FBC00 + (nLocMapForm1Index * 3)]) & 0xFFFFFF);
		nLocMapForm2Addr = 0x1E0200 + (*((uint*) &RomData[0x1FBC00 + (nLocMapForm2Index * 3)]) & 0xFFFFFF);

		DecompressData(&RomData[nLocMapForm1Addr], nLocMapForm1, 0x800);
		DecompressData(&RomData[nLocMapForm2Addr], nLocMapForm2, 0x800);

		// Store palette
		LocMapTiles.SetPalette((ushort*) &RomData[0x2DC680 + (RomData[0x2D9119 + 33 * nLocCtr] << 8)], 128);

		for(i = 0; (unsigned) i < 16; i++){
			for(j = 0; (unsigned) j < 16; j++){
				for(k = 0; k < 4; k++){
					nLocTileForm[(k << 1) + 1] = nLocMapForm1[(i << 4) + j + (k * 0x100) + 0x400];
					nLocTileForm[k << 1] = nLocMapForm1[(i << 4) + j + (k * 0x100)];

					nTileSet = nLocTileForm[(k << 1) + 1] & 0x03;
					nPaletteShift = (nLocTileForm[(k << 1) + 1] & 0x1C) >> 2;
					nFlip = nLocTileForm[(k << 1) + 1] >> 6;
					nChunk = &RawData[((nTileSet << 8) | nLocTileForm[(k << 1)]) << 5];
					LocMapTiles.Put8x8Chunk((j << 4) + ((k & 0x01) << 3), (i << 4) + ((k / 2) << 3), nChunk, nPaletteShift, nFlip);
					memcpy(&nLocMapTiles[(i << 11) + (j << 6) + ((k & 0x01) << 5) + ((k / 2) << 10)], nChunk, 32);
				}
			}
		}

		for(; (unsigned) i < 32; i++){
			for(j = 0; (unsigned) j < 16; j++){
				for(int k = 0; k < 4; k++){
					nLocTileForm[(k << 1) + 1] = nLocMapForm2[((i - 16) << 4) + j + (k * 0x100) + 0x400];
					nLocTileForm[k << 1] = nLocMapForm2[((i - 16) << 4) + j + (k * 0x100)];

					nTileSet = nLocTileForm[(k << 1) + 1] & 0x03;
					nPaletteShift = (nLocTileForm[(k << 1) + 1] & 0x1C) >> 2;
					nFlip = nLocTileForm[(k << 1) + 1] >> 6;
					nChunk = &RawData[((nTileSet << 8) | nLocTileForm[(k << 1)]) << 5];
					LocMapTiles.Put8x8Chunk((j << 4) + ((k & 0x01) << 3), (i << 4) + ((k / 2) << 3), nChunk, nPaletteShift);
					memcpy(&nLocMapTiles[(i << 11) + (j << 6) + ((k & 0x01) << 5) + ((k / 2) << 10)], nChunk, 32);
				}
			}
		}

		sprintf(sFilename, "mapdata\\LocMapTiles%03X.bmp", nLocCtr);
		LocMapTiles.WriteOutBitmap(sFilename);


		LocMap.SetPalette((ushort*) &RomData[0x2DC680 + (RomData[0x2D9119 + 33 * nLocCtr] << 8)], 128);

		nLocMap2Height = 16 << (RomData[0x2D9117 + 33 * nLocCtr] & 0x03);
		nLocMap2Width = 16 << ((RomData[0x2D9117 + 33 * nLocCtr] & 0x0C) >> 2);
		nLocMap1Height = 16 << ((RomData[0x2D9117 + 33 * nLocCtr] & 0x30) >> 4);
		nLocMap1Width = 16 << ((RomData[0x2D9117 + 33 * nLocCtr] & 0xC0) >> 6);
		if(nLocMap1Width > nLocMap2Width){
			nWidth = nLocMap1Width;
		}
		else{
			nWidth = nLocMap2Width;
		}
		if(nLocMap1Height > nLocMap2Height){
			nHeight = nLocMap1Height;
		}
		else{
			nHeight = nLocMap2Height;
		}

		LocMap.SetDim(nWidth << 4, nHeight << 4);
		LocMap.SetDefaultColor(129);

		nLocMapDataAddr = (*((ushort*) &RomData[0x2D910D + 33 * nLocCtr]) & 0x03FF);
		DecompressData(&RomData[0x19D3B0 + (*((uint*) &RomData[0x19CF90 + (3 * nLocMapDataAddr)]) & 0xFFFFFF)], LocMapData, 65536);
		nLocMapDataAddr = (*((ushort*) &RomData[0x2D910E + 33 * nLocCtr]) & 0x0FFC) >> 2;
		DecompressData(&RomData[0x19D3B0 + (*((uint*) &RomData[0x19CF90 + (3 * nLocMapDataAddr)]) & 0xFFFFFF)], LocMapData2, 65536);
		byte bLayer2OnTop = 0;

		for(i = 0; (unsigned) i < nLocMap2Height; i++){
			for(j = 0; (unsigned) j < nLocMap2Width; j++){
				nTile = LocMapData[(i * nLocMap1Width) + j];
				nTile2 = LocMapData2[(i * nLocMap2Width) + j];

				for(k = 0; k < 4; k++){
					bLayer2OnTop = ((nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 5) & 0x01) && !((nLocMapForm1[nTile + (k * 0x100) + 0x400] >> 5) & 0x01);
					if(!bLayer2OnTop){
						nPaletteShift2 = (nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 2) & 0x07;
						nFlip2 = nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 6;
						LocMap.Put8x8Chunk((j << 4) + ((k & 0x01) << 3), (i << 4) + ((k / 2) << 3), &nLocMapTiles[(((nTile2 & 0xF0) << 7) + ((nTile2 & 0x0F) << 6) + ((k & 0x01) << 5) + ((k / 2) << 10)) + 0x8000], nPaletteShift2, nFlip2);
					}
				}
			}
		}

		for(i = 0; (unsigned) i < nLocMap1Height; i++){
			for(j = 0; (unsigned) j < nLocMap1Width; j++){
				nTile = LocMapData[(i * nLocMap1Width) + j];

				for(k = 0; k < 4; k++){
					nPaletteShift = (nLocMapForm1[nTile + (k * 0x100) + 0x400] >> 2) & 0x07;
					nFlip = nLocMapForm1[nTile + (k * 0x100) + 0x400] >> 6;
					LocMap.Put8x8Chunk((j << 4) + ((k & 0x01) << 3), (i << 4) + ((k / 2) << 3), &nLocMapTiles[((nTile & 0xF0) << 7) + ((nTile & 0x0F) << 6) + ((k & 0x01) << 5) + ((k / 2) << 10)], nPaletteShift, nFlip);
				}
			}
		}

		for(i = 0; (unsigned) i < nLocMap2Height; i++){
			for(j = 0; (unsigned) j < nLocMap2Width; j++){
				nTile = LocMapData[(i * nLocMap1Width) + j];
				nTile2 = LocMapData2[(i * nLocMap2Width) + j];

				for(k = 0; k < 4; k++){
					bLayer2OnTop = ((nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 5) & 0x01) && !((nLocMapForm1[nTile + (k * 0x100) + 0x400] >> 5) & 0x01);
					if(bLayer2OnTop){
						nPaletteShift2 = (nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 2) & 0x07;
						nFlip2 = nLocMapForm2[nTile2 + (k * 0x100) + 0x400] >> 6;
						LocMap.Put8x8Chunk((j << 4) + ((k & 0x01) << 3), (i << 4) + ((k / 2) << 3), &nLocMapTiles[(((nTile2 & 0xF0) << 7) + ((nTile2 & 0x0F) << 6) + ((k & 0x01) << 5) + ((k / 2) << 10)) + 0x8000], nPaletteShift2, nFlip2);
					}
				}
			}
		}

		sprintf(sFilename, "mapdata\\LocMap%03X.bmp", nLocCtr);
		LocMap.WriteOutBitmap(sFilename);
	}//

/*		
	  byte SerpentTrenchMapData[16384];
	  DecompressData(&RomData[0x2F9F17], SerpentTrenchMapData, 16384);
	  WriteHexFile(SerpentTrenchMapData, "mapdata\\SerpentTrenchMapData.txt", 128, 16384);
	  WriteOutMap(SerpentTrenchMapData, 128, 128, "SerpentTrenchMap.bmp");
	  
	  sprintf(filename, "mapdata\\location%03X.txt", i);
	  Fout = fopen(filename, "w");
	  fprintf(Fout, "%d\n", nSize);
	  for(j = 0; j < nSize; j++){
		  fprintf(Fout, "%02X", LocationMap[j]);
	  }
	  fclose(Fout);
	  }
	  if(x){
		  sprintf(filename, "mapdata\\location%03X.bmp", i);
		  WriteOutMap(LocationMap, x, y, filename);
	  }
	  }
	  }
		
	OBBitmap Chocobo;
	ushort ChocoboPalette[16];
	Chocobo.nBitPlanes = 4;
	Fin->Seek(0x12F020, CFile::begin);
	Fin->Read(ChocoboPalette, 32);
	Chocobo.SetPalette(ChocoboPalette, 16);
	Chocobo.SetDim(128, 96);
	DecompressData(&RomData[0x2FDE4C], RawData);
	Chocobo.ProcessData(RawData, 16);
	Chocobo.WriteOutBitmap("Chocobo00.bmp");
	
	OBBitmap UnknownData;
	UnknownData.nBitPlanes = 4;
	UnknownData.SetPalette(UnknownPalette, 16);
	UnknownData.SetDim(128, 96);
	DecompressData(&RomData[0x2Fd1B9], RawData);
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown01.bmp");
	
	UnknownData.SetDim(64, 64);
	DecompressData(&RomData[0x2Fd0B7], RawData);
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown02.bmp");
	
	OBBitmap AirshipSetzer;
	ushort AirshipPaletteSetzer[16];
	AirshipSetzer.nBitPlanes = 4;
	Fin->Seek(0x12F000, CFile::begin);
	Fin->Read(AirshipPaletteSetzer, 32);
	AirshipSetzer.SetPalette(AirshipPaletteSetzer, 16);
	AirshipSetzer.SetDim(128, 96);
	DecompressData(&RomData[0x2EC902], RawData);
	AirshipSetzer.ProcessData(RawData, 16);
	AirshipSetzer.WriteOutBitmap("AirshipSetzer.bmp");
	
	UnknownData.SetDim(128, 128);
	DecompressData(&RomData[0x2EB490], RawData);
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown03.bmp");
	
	UnknownData.SetDim(128, 144);
	UnknownData.nBitPlanes = 1;
	DecompressData(&RomData[0x2FB831], RawData);
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown04.bmp");
	
	DecompressData(&RomData[0x2FC824], RawData);
	Chocobo.ProcessData(RawData, 16);
	Chocobo.WriteOutBitmap("Chocobo01.bmp");
	
	ulong BattleBGPtr[75];
	memset(BattleBGPtr, 0, 300);
	Fin->Seek(0x271850, CFile::begin);
	for(i = 0; i < 75; i++){
		Fin->Read(&BattleBGPtr[i], 3);
	}
	
	UnknownData.SetDim(128, 64);
	DecompressData(&RomData[0x26E045], RawData);
	UnknownData.nBitPlanes = 4;
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown05.bmp");

	Waveform BRR;
	uint nRomPtr;
	for(i = 0; i < 63; i++){
		nRomPtr = (RomData[0x053E5F + (i * 3)] | (RomData[0x053E60 + (i * 3)] << 8) | (RomData[0x053E61 + (i * 3)] << 16)) - 0xBFFE00;
		BRR.SetDataSize((((RomData[nRomPtr] | (RomData[nRomPtr + 1] << 8)) - 2) << 5) / 9);
		BRR.SetSampleRate(8000);
		BRR.DecodeBRR(&RomData[nRomPtr]);
		sprintf(filename, "brrs\\sample%02X.wav", i);
		BRR.WriteOutWaveform(filename);
	}
*/	
/*
	UnknownData.SetDim(64, 128);
	ulong nLength = DecompressData(&RomData[0x26CFC0], RawData);
	UnknownData.nBitPlanes = 4;
	UnknownData.ProcessData(RawData, 16);
	UnknownData.WriteOutBitmap("Unknown07.bmp");
*/
	/*	Fout = fopen("brr offsets.txt", "w");
	// Find BRRs
	j = 0;
	byte bLoopBitSet;
	byte bBadLoop;
	uint nPrevBRR = 0;
	ushort nPrevBRRLen = 0;
	ushort nCurBRRLen;
	byte nBRRs = 0;
	uint nBRRArray[63];
	byte nBRRMatch[63];

	for(i = 0x053E5F; i < 0x085E7A; i++){
		bBadLoop = 0;
		j = 2;
		nCurBRRLen = RomData[i] | (RomData[i + 1] << 8);
		if(nCurBRRLen && !(nCurBRRLen % 9)){
			while(i + j < 0x085E7A && (RomData[i + j] >> 4) < 13){
				if(j == 2){
					bLoopBitSet = RomData[i + j] & 0x02;
				}
				else if((RomData[i + j] & 0x02) != bLoopBitSet){
					bBadLoop = 1;
					break;
				}

				j += 9;

				if(RomData[i + j] & 0x01){
					j += 9;
					break;
				}
			}
			j -= 2;
			if(j == nCurBRRLen && !bBadLoop && j > 1000){
				if(i + 1 > nPrevBRR + nPrevBRRLen || (i - nPrevBRR) % 9){
					nBRRs++;
					for(int k = 0; nBRRArray[k] != i && k < 63; k++){
					}
					if(k != 63){
						nBRRMatch[k] = 1;
					}
					nPrevBRR = i;
					nPrevBRRLen = j;
				}
			}
		}
	}
	fclose(Fout);*/


	
	
/*	
	// BEGIN: Fin->Read global palettes.
	Fin->Seek(0x268200, CFile::begin);
	Fin->Read(characterPalettes, 256);
	
	Fin->Seek(0x2D5A60, CFile::begin);
	Fin->Read(portraitPalettes, 608);
	
	Fin->Seek(0x00D03A, CFile::begin);
	Fin->Read(characterGraphic, 12 * 40);
	
	Fin->Seek(0x0491C0, CFile::begin);
	Fin->Read(charWidths, 128);
	
	Fin->Seek(0x0492C0, CFile::begin);
	Fin->Read(&variableFont, 11 * 128 * sizeof(ushort));
	
	Fin->Seek(0x0489C0, CFile::begin);
	Fin->Read(&fixedFont, 16 * 128);
	
	i = 0;
	
	// BEGIN: Fin->Read character data.
	Fin->Seek(0x2D7EA0, CFile::begin);
	for(int n = 0; n < 64; n++){
		Fin->Read(&characters[n], 22);
	}
	
	Fin->Seek(0x047AC0, CFile::begin);
	for(n = 0; n < 64; n++){
		Fin->Read(characters[n].name, 6);
	}
	
	//	WriteOutCharacterData();
	
	for(n = 0; n < 32; n++){
		ushort wLowPtr = 0;
		Fin->Seek(0x00D2F2 + n * 2, CFile::begin);
		Fin->Read(&wLowPtr, 2);
		
		byte bHiPtr = 0;
		Fin->Seek(0x00D43C + n * 2, CFile::begin);
		Fin->Read(&bHiPtr, 1);
		
		ulong dwPtr = bHiPtr << 16 | wLowPtr;
		Fin->Seek(dwPtr, CFile::begin);
		Fin->Read(characters[n].gfx, 0x16A0);
		
		Fin->Seek(0x2D1F00 + (n * 800), CFile::begin);
		Fin->Read(characters[n].portrait, 800);
	}
	
	Fin->Seek(0x02D02B, CFile::begin);
	for(n = 0; n < 32; n++){
		Fin->Read(&characters[n].palette, 1);
	}
	
	WriteOutPortraits();
	
	Fin->Seek(0x26F6A0, CFile::begin);
	Fin->Read(hpProgression, 98);
	Fin->Read(mpProgression, 98);
	
	// END: Fin->Read character data.
	
	// Fin->Read monster formation data.
	Fin->Seek(0x0F6400, CFile::begin);
	for(n = 0; n < 576; n++){
		Fin->Read(&monsterFormations[n], 15);
	}
	
	// Fin->Read monster formation auxiliary data.
	Fin->Seek(0x0F5B00, CFile::begin);
	for(n = 0; n < 576; n++){
		Fin->Read(&monsterFormations[n].aux, 4);
	}
	
	// Fin->Read monster data.
	Fin->Seek(0x0F0200, CFile::begin);
	for(n = 0; n < 384; n++){
		Fin->Read(&monsters[n].strength, 32);
	}
	
	Fin->Seek(0x0FC250, CFile::begin);
	for(n = 0; n < 384; n++){
		Fin->Read(&monsters[n].name, 10);
	}
	
	Fin->Seek(0x0FD2D0, CFile::begin);
	for(n = 0; n < 384; n++){
		Fin->Read(&monsters[n].attackName, 10);
	}
	
	Fin->Seek(0x0F3200, CFile::begin);
	for(n = 0; n < 384; n++){
		Fin->Read(&monsters[n].itemsStolen, 4);
	}
	
	Fin->Seek(0x048140, CFile::begin);
	Fin->Read(morphPackages, 26 * 4);
	
	//	WriteOutMonsterData();
	
	for(n = 0; n < 13; n++){
		int m = 0;
		byte bDone = 0;
		
		Fin->Seek(0x02D21A + n * 2, CFile::begin);
		ushort wPtr, w1;
		byte b1, b2;
		Fin->Read(&wPtr, 2);
		
		Fin->Seek(0x020200 + wPtr, CFile::begin);
		Fin->Read(&w1, 2);
		Fin->Read(&b1, 1);
		Fin->Read(&b2, 1);
		monsterTemplates[n][m].SetUnknown(w1);
		monsterTemplates[n][m].SetWidth(b1);
		monsterTemplates[n][m].SetHeight(b2);
		m++;
		Fin->Read(&w1, 2);
		Fin->Read(&b1, 1);
		Fin->Read(&b2, 1);
		
		while(w1 && m < 6){
			monsterTemplates[n][m].SetUnknown(w1);
			monsterTemplates[n][m].SetWidth(b1);
			monsterTemplates[n][m].SetHeight(b2);
			m++;
			Fin->Read(&w1, 2);
			Fin->Read(&b1, 1);
			Fin->Read(&b2, 1);
		}
		
		monsterTemplateCount[n] = m;
	}
	
	// Fin->Read monster palettes.
	Fin->Seek(0x127A20, CFile::begin);
	Fin->Read(monsterPalettes, 0x3000);
	
	// Fin->Read monster visuals.
	Fin->Seek(0x127200, CFile::begin);
	for(n = 0; n < 384; n++){
		Fin->Read(&monsterVisuals[n], 5);
	}
	
	// Fin->Read monster compositions.
	ushort wPtr;
	Fin->Seek(0x12AA20, CFile::begin);
	Fin->Read(&wPtr, 2);
	Fin->Seek(0x120200 + wPtr, CFile::begin);
	for(n = 0; n < 128; n++){
		monsterCompositions[n].length = 8;
		Fin->Read(monsterCompositions[n].mask, 8);
	}
	
	Fin->Seek(0x12AA22, CFile::begin);
	Fin->Read(&wPtr, 2);
	Fin->Seek(0x120200 + wPtr, CFile::begin);
	for(n = 128; n < 176; n++){
		monsterCompositions[n].length = 32;
		Fin->Read(monsterCompositions[n].mask, 32);
	}
	
	// Fin->Read monster gfx.
	Fin->Seek(0x297200, CFile::begin);
	Fin->Read(monsterGraphics, 0x38000);
	
	//	WriteOutMonsters();
	
	// Fin->Read item data.
	for(n = 0; n < 256; n++){
		memcpy(items[n].name, &RomData[0x12B500 + 13 * n], 13);
		memcpy(items[n].props, &RomData[0x185200 + 30 * n], 30);
		uint nAddress = 0x2D6600 + *((ushort*) &RomData[0x2D7CA0 + (n << 1)]);
		if(RomData[nAddress] != 0){
			for(i = 0; RomData[nAddress + i] != 0; i++){
				items[n].description[i] = RomData[nAddress + i];
			}
		}
		items[n].description[i] = 0;
	}
	WriteOutItemData();
	
	// Fin->Read shop data.
	Fin->Seek(0x047CC0, CFile::begin);
	Fin->Read(shops, 9 * 128);
	
	// Fin->Read esper data.
	Fin->Seek(0x26F8E1, CFile::begin);
	for(n = 0; n < 27; n++){
		Fin->Read(espers[n].name, 8);
	}
	
	Fin->Seek(0x187000, CFile::begin);
	for(n = 0; n < 27; n++){
		Fin->Read(&espers[n].spell1Rate, 11);
	}
	
	// Fin->Read magic data.
	Fin->Seek(0x26F767, CFile::begin);
	for(n = 0; n < 0x36; n++){
		Fin->Read(magic[n].name, 7);
	}
	
	for(n = 0x36; n < 0x51; n++){
		Fin->Read(magic[n].name, 8);
	}
	
	for(n = 0x51; n < 0x55; n++){
		Fin->Read(magic[n].name, 10);
	}
	
	Fin->Seek(0x0F3E40, CFile::begin);
	for(n = 0x56; n < 0x5E; n++){
		Fin->Read(magic[n - 1].name, 12);
	}
	
	Fin->Seek(0x26FA31, CFile::begin);
	for(n = 0x5E; n < 0x100; n++){
		Fin->Read(magic[n - 1].name, 10);
	}
	
	Fin->Seek(0x046CC0, CFile::begin);
	for(n = 0; n < 256; n++){
		Fin->Read(&magic[n].target, 14);
	}
	
	// Fin->Read the battle command data.
	Fin->Seek(0x18D0A0, CFile::begin);
	for(n = 0; n < 32; n++){
		Fin->Read(battleCommands[n].name, 7);
	}
	
	//Menu Screen Graphics
	Fin->Seek(0x2D0200, CFile::begin);
	Fin->Read(MenuScreenGraphics, 7168);
	Fin->Seek(0x2D1E00, CFile::begin);
	Fin->Read(MenuScreenPalettes, 256);
	//	WriteOutMenuScreens();
	*/

	// Battle backgrounds

/*		OBBitmap LocMap;

		LocMap.SetDim(32, 19);
		LocMap.SetPaletteGrey();
		for(byte dy = 0; dy < 19; dy++){
			for(byte dx = 0; dx < 32; dx++){
				LocMap.Data[(19 - dy - 1) * 32 + dx] = RawData[dy * 64 + (dx << 1)];
			}
		}
		sprintf(filename, "battlebackgrounds\\BatBGMap%02X.bmp", i);
		LocMap.WriteOutBitmap(filename);

		sprintf(filename, "battlebackgrounds\\BatBG%02X.txt", i);
		WriteHexFile(RawData, filename, 64, 19 << 6);*/

/*
	//Event Code
	Fin->Seek(0x029B1D + 0x0A0200, CFile::begin);
	Fin->Read(&event, 8192);
	
	// Find out if the dialogue routine has been changed.
	byte opcode = 0;
	Fin->Seek(0x0081BF, CFile::begin);
	Fin->Read(&opcode, 1);
	
	if(opcode == 0xA9){
		ushort wPageBound = 0;
		Fin->Seek(0xCE800, CFile::begin);
		Fin->Read(&wPageBound, 2);
		
		byte bTemp[1024];
		
		for(ulong i = 0xCE802; i < 0xD01FE; i += 2){
			int nPos = 0;
			
			DialogueItem* pDlg = new DialogueItem;
			dialogue.Add(pDlg);
			
			ushort wBegin = 0, wEnd = 0;
			Fin->Seek(i, CFile::begin);
			Fin->Read(&wBegin, 2);
			Fin->Read(&wEnd, 2);
			
			ulong dwBegin = wBegin, dwEnd = wEnd;
			
			if((i - 0xCE802) / 2 + 1 >= wPageBound) dwEnd += 0x10000;
			if((i - 0xCE802) / 2 >= wPageBound) dwBegin += 0x10000;
			
			ulong dwLength = dwEnd - dwBegin;
			
			if(dwLength){
				byte bChar;
				Fin->Seek(dwBegin + 0xD0200, CFile::begin);
				Fin->Read(&bChar, 1);
				
				while(bChar){
					bTemp[nPos++] = bChar;
					Fin->Read(&bChar, 1);
				}
				
				bTemp[nPos++] = 0x00;
				
				pDlg->text = new byte[nPos];
				pDlg->length = nPos;
				CopyMemory(pDlg->text, bTemp, nPos);
			}
			else{
				pDlg->text = 0;
				pDlg->length = 0;
			}
		}
	}
	else if(opcode == 0x7B){
		ushort ptrs[10];
		byte pages[10];
		
		Fin->Seek(0x00E180, CFile::begin);
		i = 0;
		
		Fin->Read(&ptrs[i], 2);
		Fin->Read(&pages[i], 1);
		
		while(ptrs[i] != 0xFFFF){
			i++;
			Fin->Read(&ptrs[i], 2);
			Fin->Read(&pages[i], 1);
		}
		
		i++;
		int nNumPages = i;
		
		ushort wItems;
		Fin->Read(&wItems, 2);
		dialogue.SetSize(wItems);
		
		ushort* dlgPtrs = new ushort[wItems];
		
		Fin->Seek(0x0CE802, CFile::begin);
		Fin->Read(dlgPtrs, dialogue.GetSize() * sizeof(ushort));
		
		byte bTemp[1024];
		
		int i = 0;
		for(int n = 0; n < nNumPages; n++){
			while(i < ptrs[n] && i < dialogue.GetSize()){
				int nPos = 0;
				
				DialogueItem* pDlg = new DialogueItem;
				dialogue[i] = pDlg;
				
				if((i < dialogue.GetSize()) && (dlgPtrs[i] != dlgPtrs[i + 1])){
					byte bChar;
					Fin->Seek(pages[n] << 16 | dlgPtrs[i], CFile::begin);
					Fin->Read(&bChar, 1);
					
					while(bChar){
						bTemp[nPos++] = bChar;
						Fin->Read(&bChar, 1);
					}
					
					bTemp[nPos++] = 0x00;
					
					pDlg->text = new byte[nPos];
					pDlg->length = nPos;
					CopyMemory(pDlg->text, bTemp, nPos);
				}
				else{
					pDlg->text = 0;
					pDlg->length = 0;
				}
				
				i++;
			}
		}
		
		delete [] dlgPtrs;
	}
	
	m_bValid = 1;
	
	ReleaseFile(Fin, 0);
	SetModifiedFlag(0);     // start off with unmodified
	
	Fin = 0;

	cout << "Completed all operations." << endl;
*/	
	return 1;
}




void CFF3Doc::WriteOutMonsters(){
	char filename[64];
	MonsterVisuals *pVis;
	MonsterComposition *pComp;
	int offset = 0;
	OBBitmap Monster;

	for(int nMonstNum = 0; nMonstNum < 384; nMonstNum++){
		pVis = &monsterVisuals[nMonstNum];
		offset = 0;

		Monster.SetPalette(&monsterPalettes[(((pVis->palette | (pVis->extra << 8)) & 0x03FF) * 8)], 16);
		
		if(pVis->extra & 0x80){
			pComp = &monsterCompositions[pVis->composition + 128];
		}
		else{
			pComp = &monsterCompositions[pVis->composition];
		}
		
		if(pVis->Is3Bitplane()){
			Monster.nBitPlanes = 3;
		}
		else{
			Monster.nBitPlanes = 4;
		}

		if(pComp->length == 8){			//64x64
			Monster.SetDim(64, 64);
			Monster.SetDefaultColor(16);

			for(int y = 0; y < 8; y++){
				int x = 0;
				
				for(int bit = 7; bit >= 0; bit--, x += 8){
					if(pComp->mask[y] & (1 << bit)){
						Monster.Put8x8Chunk(x, y * 8, &monsterGraphics[(pVis->gfx & 0x7FFF) * 8 + offset]);
						offset += (Monster.nBitPlanes << 3);
					}
				}
			}
		}
		else if(pComp->length == 32){	//128x128
			Monster.SetDim(128, 128);
			Monster.SetDefaultColor(16);
			
			for(int y = 0; y < 16; y++){
				int x = 0;
				
				for(int bit = 15; bit >= 0; bit--, x += 8){
					ushort wMask = pComp->mask[y * 2 + 1] | (pComp->mask[y * 2] << 8);
					
					if(wMask & (1 << bit)){
						Monster.Put8x8Chunk(x, y * 8, &monsterGraphics[(pVis->gfx & 0x7FFF) * 8 + offset]);
						offset += (Monster.nBitPlanes << 3);
					}
				}
			}
		}
		
		Monster.ResetTransparency(0, 16);

		char sMonsterName[11];
		TranslateString(monsters[nMonstNum].name, sMonsterName, 10);
		sprintf(filename, "monstergraphics\\%03X %s.bmp", nMonstNum, sMonsterName);
		Monster.WriteOutBitmap(filename);
	}
}



void TranslateString(byte *String, char *sNewStr, byte nStrLen){
	int i;
	byte bLastPosFound = 0;

	for(i = nStrLen - 1; i > -1; i--){
		if(String[i] > 153 && String[i] < 180){
			sNewStr[i] = String[i] - 57;
			bLastPosFound = 1;
		}
		else if(String[i] > 127 && String[i] < 154){
			sNewStr[i] = String[i] - 63;
			bLastPosFound = 1;
		}
		else if(String[i] > 179 && String[i] < 190){
			sNewStr[i] = String[i] - 132;
			bLastPosFound = 1;
		}
		else if(String[i] == 191){
			sNewStr[i] = '?';
			bLastPosFound = 1;
		}
		else if(String[i] == 196){
			sNewStr[i] = '-';
			bLastPosFound = 1;
		}
		else if(String[i] == 197){
			sNewStr[i] = '.';
		}
		else{
			if(!bLastPosFound){
				sNewStr[i] = '\0';
			}
			else{
				sNewStr[i] = ' ';
			}
		}
	}

	sNewStr[nStrLen] = '\0';
}



void TranslateString(byte *String, string *sNewStr, byte nStrLen){
	int i;
	byte bLastPosFound = 0;

	for(i = nStrLen - 1; i > -1; i--){
		if(String[i] == 0x01){
			*sNewStr = "\\n" + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] > 127 && String[i] < 154){
			*sNewStr = (char) (String[i] - 63) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] > 153 && String[i] < 180){
			*sNewStr = (char) (String[i] - 57) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] > 179 && String[i] < 190){
			*sNewStr = (char) (String[i] - 132) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 190){
			*sNewStr = '!' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 191){
			*sNewStr = '?' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 192){
			*sNewStr = '/' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 195){
			*sNewStr = '\'' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 196){
			*sNewStr = '-' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 197){
			*sNewStr = '.' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 198){
			*sNewStr = ',' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 205){
			*sNewStr = '%' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 194 || String[i] == 211 || (String[i] > 215 && String[i] < 232) || String[i] == 0){
		}
		//This case separates legitimate spaces from unmapped characters
		else if(String[i] > 253){
			if(bLastPosFound){
				*sNewStr = ' ' + *sNewStr;
			}
		}
		else{
			if(bLastPosFound){
				*sNewStr = ' ' + *sNewStr;
			}
		}
	}
}



void TranslateString2(byte *String, string *sNewStr, byte nStrLen){
	int i;
	byte bLastPosFound = 0;

	for(i = nStrLen - 1; i > -1; i--){
		if(String[i] > 0x19 && String[i] < 0x3a){
			*sNewStr = (char) (String[i] + 0x21) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] > 0x39 && String[i] < 0x54){
			*sNewStr = (char) (String[i] + 0x27) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] > 0x53 && String[i] < 0x5E){
			*sNewStr = (char) (String[i] - 0x24) + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 0x60){
			*sNewStr = '/' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 0x63){
			*sNewStr = '\'' + *sNewStr;
			bLastPosFound = 1;
		}
		else if(String[i] == 0x65){
			*sNewStr = '.' + *sNewStr;
			bLastPosFound = 1;
		}
		//This case separates legitimate spaces from unmapped characters
		else if(String[i] == 0x7F){
			if(bLastPosFound){
				*sNewStr = ' ' + *sNewStr;
			}
		}
		else{
			if(bLastPosFound){
				*sNewStr = '~' + *sNewStr;
			}
		}
	}
}



void CFF3Doc::WriteOutMonsterData(){
	char sMonsterName[11];
	int i;
	FILE *Fout;
	char *sFilename = "monsterdata.txt";
	char sMonsterAttack[11];

	Fout = fopen(sFilename, "w");
	fprintf(Fout, "Index,Name,Attack,Steal,StealRare,Drop,DropRare,Strength,Speed,Hit,Evade,MagicEvade,Defense,");
	fprintf(Fout, "MagicDefense,Magic,HP,MP,XP,Gold,Lvl,Morph,Flags1,Flags2,ImmuneStatus1,ImmuneStatus2,Unknown3,Absorb,NoEffect,");
	fprintf(Fout, "Weak,AttackType,Condition1,Condition2,Condition3,Condition4,SpecialAttack\n");
	for(i = 0; i < 384; i++){
		TranslateString(monsters[i].name, sMonsterName, 10);
		TranslateString(monsters[i].attackName, sMonsterAttack, 10);
		fprintf(Fout, "%d,%s,%s,%d,%d,", i, sMonsterName, sMonsterAttack, monsters[i].itemsStolen[0], monsters[i].itemsStolen[1]);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].itemsDropped[0], monsters[i].itemsDropped[1], monsters[i].strength, monsters[i].speed);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].hitRate, monsters[i].evadeRate, monsters[i].magBlockRate, monsters[i].defense);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].magicDefense, monsters[i].magicPower, monsters[i].hp, monsters[i].mp);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].exp, monsters[i].gold, monsters[i].level, monsters[i].morph);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].flags1, monsters[i].flags2, monsters[i].ImmuneStatus1, monsters[i].ImmuneStatus2);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].ImmuneStatus3, monsters[i].elemAbsorb, monsters[i].elemNoEffect, monsters[i].elemWeak);
		fprintf(Fout, "%d,%d,%d,%d,", monsters[i].attackType, monsters[i].conditions[0], monsters[i].conditions[1], monsters[i].conditions[2]);
		fprintf(Fout, "%d,%d\n", monsters[i].conditions[3], monsters[i].specialAttack);
	}
	fclose(Fout);
}



void CFF3Doc::WriteOutPortraits(){
	char filename[64];
	int i;
	int j;
	OBBitmap Portrait;

	Portrait.SetDim(40, 40);
	Portrait.nBitPlanes = 4;
	
	for(i = 0; i < 19; i++){
		Portrait.SetPalette(&portraitPalettes[i << 4], 16);

		Portrait.Put8x8Chunk(0, 0, &characters[i].portrait[(0 * 20) + (0 << 2)]);
		Portrait.Put8x8Chunk(8, 0, &characters[i].portrait[(0 * 20) + (8 << 2)]);
		Portrait.Put8x8Chunk(16, 0, &characters[i].portrait[(0 * 20) + (16 << 2)]);
		Portrait.Put8x8Chunk(24, 0, &characters[i].portrait[(0 * 20) + (24 << 2)]);
		Portrait.Put8x8Chunk(32, 0, &characters[i].portrait[(8 * 20) + (24 << 2)]);

		Portrait.Put8x8Chunk(0, 8, &characters[i].portrait[(24 * 20) + (8 << 2)]);
		Portrait.Put8x8Chunk(8, 8, &characters[i].portrait[(24 * 20) + (16 << 2)]);
		Portrait.Put8x8Chunk(16, 8, &characters[i].portrait[(24 * 20) + (24 << 2)]);
		Portrait.Put8x8Chunk(24, 8, &characters[i].portrait[(24 * 20) + (32 << 2)]);
		Portrait.Put8x8Chunk(32, 8, &characters[i].portrait[(8 * 20) + (32 << 2)]);
		
		Portrait.Put8x8Chunk(0, 16, &characters[i].portrait[(0 * 20) + (32 << 2)]);
		Portrait.Put8x8Chunk(8, 16, &characters[i].portrait[(8 * 20) + (0 << 2)]);
		Portrait.Put8x8Chunk(16, 16, &characters[i].portrait[(8 * 20) + (8 << 2)]);
		Portrait.Put8x8Chunk(24, 16, &characters[i].portrait[(8 * 20) + (16 << 2)]);
		Portrait.Put8x8Chunk(32, 16, &characters[i].portrait[(16 * 20) + (0 << 2)]);
		
		Portrait.Put8x8Chunk(0, 24, &characters[i].portrait[(32 * 20) + (0 << 2)]);
		Portrait.Put8x8Chunk(8, 24, &characters[i].portrait[(32 * 20) + (8 << 2)]);
		Portrait.Put8x8Chunk(16, 24, &characters[i].portrait[(32 * 20) + (16 << 2)]);
		Portrait.Put8x8Chunk(24, 24, &characters[i].portrait[(32 * 20) + (24 << 2)]);
		Portrait.Put8x8Chunk(32, 24, &characters[i].portrait[(16 * 20) + (8 << 2)]);
		
		Portrait.Put8x8Chunk(0, 32, &characters[i].portrait[(16 * 20) + (24 << 2)]);
		Portrait.Put8x8Chunk(8, 32, &characters[i].portrait[(16 * 20) + (32 << 2)]);
		Portrait.Put8x8Chunk(16, 32, &characters[i].portrait[(24 * 20) + (0 << 2)]);
		Portrait.Put8x8Chunk(24, 32, &characters[i].portrait[(32 * 20) + (32 << 2)]);
		Portrait.Put8x8Chunk(32, 32, &characters[i].portrait[(16 * 20) + (16 << 2)]);

		char sName[7];
		TranslateString(characters[i].name, sName, 6);
		for(j = 0; j < 6; j++){
			if(sName[j] == '?'){
				sName[j] = '-';
			}
		}
		
		sprintf(filename, "portraits\\%02X %s.bmp", i, sName);
		Portrait.WriteOutBitmap(filename);
	}
}


void CFF3Doc::WriteOutCharacterData(){
	char sCharacterName[7];
	int i;
	FILE *Fout;
	char *sFilename = "characterdata.txt";

	Fout = fopen(sFilename, "w");
	fprintf(Fout, "Index,Name,InitHP,InitMP,Command1,Command2,Command3,Command4,Vigor,Speed,Stamina,Magic,Attack,Defense,");
	fprintf(Fout, "MagDef,Evade,MagEvade,Weapon,Shield,Hat,Armor,Relic1,Relic2,Level\n");
	for(i = 0; i < 64; i++){
		TranslateString(characters[i].name, sCharacterName, 6);
		fprintf(Fout, "%d,%s,%d,%d,", i, sCharacterName, characters[i].initialHP, characters[i].initialMP);
		fprintf(Fout, "%d,%d,%d,%d,", characters[i].battleCmds[0], characters[i].battleCmds[1], characters[i].battleCmds[2], characters[i].battleCmds[3]);
		fprintf(Fout, "%d,%d,%d,%d,", characters[i].vigor, characters[i].speed, characters[i].stamina, characters[i].magicPower);
		fprintf(Fout, "%d,%d,%d,%d,", characters[i].battlePower, characters[i].defense, characters[i].magicDefense, characters[i].evade);
		fprintf(Fout, "%d,%d,%d,%d,", characters[i].magicBlock, characters[i].weapon, characters[i].shield, characters[i].hat);
		fprintf(Fout, "%d,%d,%d,%d\n", characters[i].armor, characters[i].relic1, characters[i].relic2, characters[i].level);
	}
	fclose(Fout);
}



void CFF3Doc::WriteOutMenuScreens(){
	char filename[64];
	int i;
	OBBitmap MenuScreen;

	MenuScreen.SetDim(32, 56);
	MenuScreen.nBitPlanes = 4;

	for(i = 0; i < 8; i++){
		MenuScreen.SetPalette(&MenuScreenPalettes[i << 4], 16);
		MenuScreen.ProcessData(&MenuScreenGraphics[i * 896], 16);
		sprintf(filename, "menuscreens\\%02X.bmp", i);
		MenuScreen.WriteOutBitmap(filename);
	}
}


void CFF3Doc::WriteOutMiniMap(){
	char filename[64];
	int i;
	OBBitmap MiniMap;

	MiniMap.SetDim(64, 64);
	MiniMap.nBitPlanes = 4;
	MiniMap.SetPaletteGrey(9);

	for(i = 0; i < 2; i++){
		int y = 0;
		for(int x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, y, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 8;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, 16, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 16;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, 8, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 24;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, y, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 32;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, y, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 40;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, 48, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 48;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, 40, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		y = 56;
		for(x = 0; x < 64; x += 8){
			MiniMap.Put8x8Chunk(x, y, &worldData[i].miniMap[(y << 5) + (x << 2)]);
		}
		
		sprintf(filename, "minimap\\%02X.bmp", i);
		MiniMap.WriteOutBitmap(filename);
	}
}



void CFF3Doc::WriteLocation(byte *Data, uint width, uint height, ushort *Palette, char *filename){
	OBBitmap Loc;
	
	Loc.SetDim(width, height);
	Loc.SetPaletteGrey();
	Loc.nBitPlanes = 4;
	Loc.SetPalette(Palette, 16);
	Loc.ProcessData(Data, 16);
	Loc.WriteOutBitmap(filename);	
}


void CFF3Doc::WriteOutItemData(){
	string sItemName;
	string sItemDesc;
	int i;
	ofstream Fout;

	Fout.open("itemdata.txt", ios_base::out);
	Fout << "Index\tName\tDescription\t";
	Fout << "Type\tEquipChars\tSpellRate\tSpellLearned\tFieldEffects\tCond1\tCond2\tCond3Equip\t";
	Fout << "Status1\tBattle1\tStatus2\tBattle2\tBattle3\tTarget\tElements\tVigor/Speed\t";
	Fout << "Stam/MagPwr\tAttack\tProps\tProp20\tProp21\tProp22\tProp23\tProp24\tCond2Equip\tEvade/MBlock\t";
	Fout << "Special\tPrice\n";
	for(i = 0; i < 256; i++){
		sItemName.clear();
		sItemDesc.clear();
		TranslateString(items[i].name, &sItemName, 13);
		TranslateString(items[i].description, &sItemDesc, 64);
		Fout << i << "\t" << sItemName << "\t" << sItemDesc << "\t";
		Fout << (ushort) items[i].props[0] << "\t" << *((ushort*) &items[i].props[1]) << "\t" << (ushort) items[i].props[3] << "\t" << (ushort) items[i].props[4] << "\t";
		Fout << (ushort) items[i].props[5] << "\t" << (ushort) items[i].props[6] << "\t" << (ushort) items[i].props[7] << "\t" << (ushort) items[i].props[8] << "\t";
		Fout << (ushort) items[i].props[9] << "\t" << (ushort) items[i].props[10] << "\t" << (ushort) items[i].props[11] << "\t" << (ushort) items[i].props[12] << "\t";
		Fout << (ushort) items[i].props[13] << "\t" << (ushort) items[i].props[14] << "\t" << (ushort) items[i].props[15] << "\t" << (ushort) items[i].props[16] << "\t";
		Fout << (ushort) items[i].props[17] << "\t" << (ushort) items[i].props[18] << "\t" << (ushort) items[i].props[19] << "\t" << (ushort) items[i].props[20] << "\t";
		Fout << (ushort) items[i].props[21] << "\t" << (ushort) items[i].props[22] << "\t" << (ushort) items[i].props[23] << "\t" << (ushort) items[i].props[24] << "\t";
		Fout << (ushort) items[i].props[25] << "\t" << (ushort) items[i].props[26] << "\t" << (ushort) items[i].props[27] << "\t" << *((ushort*) &items[i].props[28]) << "\n";
	}
	Fout.close();
}


void CFF3Doc::WriteOutEsperData(){
	int i;
	string sEsperDescription;
	string sEsperName;
	byte nStrLen;
	uint nDescriptionOffset = 0x0F3B40;
	ofstream Fout;

	Fout.open("esperdata.txt");

	Fout << "Index\tName\tDescription\tSpell1Rate\tSpell1\tSpell2Rate\tSpell2\tSpell3Rate\tSpell3\tSpell4Rate\tSpell4\tSpell5Rate\tSpell5\tBonus\n";
	for(i = 0; i < 27; i++){
		Fout << i << "\t";
		sEsperName = "";
		TranslateString(&RomData[0x26F8E1 + (i << 3)], &sEsperName, 8);
		Fout << sEsperName << "\t";
		
		nStrLen = 0;
		sEsperDescription = "";
		while(RomData[nDescriptionOffset++] != 0x00){
			nStrLen++;
		}
		TranslateString(&RomData[nDescriptionOffset - (nStrLen + 1)], &sEsperDescription, nStrLen);
		Fout << sEsperDescription << "\t";

		Fout << (ushort) RomData[0x187000 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187001 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187002 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187003 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187004 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187005 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187006 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187007 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187008 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x187009 + (i * 11)] << "\t";
		Fout << (ushort) RomData[0x18700A + (i * 11)] << "\n";
	}
}


/*
void CFF3Doc::WriteOutEspers(){
	char filename[64];
	MonsterVisuals EsperVisuals[32];
	int offset = 0;
	OBBitmap Esper;
	int i;

	for(i = 0; i < 32; i++){
		memcpy(&EsperVisuals[i], &RomData[0x127980 + (i * 5)], 5);
		offset = 0;

		//Monster.SetPalette(&monsterPalettes[(((EsperVisuals[i].palette | (EsperVisuals[i].extra << 8)) & 0x03FF) * 8)], 16);
		
		//&monsterCompositions[EsperVisuals[i].composition + (EsperVisuals[i].extra & 0x80)];
		
		if(EsperVisuals[i].Is3Bitplane()){
			Monster.nBitPlanes = 3;
		}
		else{
			Monster.nBitPlanes = 4;
		}

//		if(pComp->length == 8){			//64x64
			Monster.SetDim(64, 64);
			Monster.SetDefaultColor(16);

			for(int y = 0; y < 8; y++){
				int x = 0;
				
				for(int bit = 7; bit >= 0; bit--, x += 8){
					if(pComp->mask[y] & (1 << bit)){
						Monster.Put8x8Chunk(x, y * 8, &monsterGraphics[(EsperVisuals[i].gfx & 0x7FFF) * 8 + offset]);
						offset += (Monster.nBitPlanes << 3);
					}
				}
			}
		}
		else if(pComp->length == 32){	//128x128
			Monster.SetDim(128, 128);
			Monster.SetDefaultColor(16);
			
			for(int y = 0; y < 16; y++){
				int x = 0;
				
				for(int bit = 15; bit >= 0; bit--, x += 8){
					ushort wMask = pComp->mask[y * 2 + 1] | (pComp->mask[y * 2] << 8);
					
					if(wMask & (1 << bit)){
						Monster.Put8x8Chunk(x, y * 8, &monsterGraphics[(EsperVisuals[i].gfx & 0x7FFF) * 8 + offset]);
						offset += (Monster.nBitPlanes << 3);
					}
				}
			}
		}
		
		Monster.ResetTransparency(0, 16);

		char sMonsterName[11];
		TranslateString(monsters[nMonstNum].name, sMonsterName, 10);
		sprintf(filename, "monstergraphics\\%03X %s.bmp", nMonstNum, sMonsterName);
		Monster.WriteOutBitmap(filename);
	}
}*/



void CFF3Doc::WriteLocPropData(){
	int i;
	ofstream Fout;
	uint nOut = 0;

	Fout.open("locpropdata.txt");

	Fout << "Index\t";
	Fout << "LocName\t";
	Fout << "Byte01\t";
	Fout << "Byte02\t";
	Fout << "Byte03\t";
	Fout << "TileProp\t";
	Fout << "Byte05\t";
	Fout << "Byte06\t";
	Fout << "Tileset1\t";
	Fout << "Tileset2\t";
	Fout << "Tileset3\t";
	Fout << "Tileset4\t";
	Fout << "Byte10\t";
	Fout << "TileForm1\t";
	Fout << "TileForm2\t";
	Fout << "Map1\t";
	Fout << "Map2\t";
	Fout << "Map3\t";
	Fout << "Byte17\t";
	Fout << "Byte18\t";
	Fout << "Byte19\t";
	Fout << "Byte20\t";
	Fout << "Byte21\t";
	Fout << "Scroll\t";
	Fout << "L2Height\t";
	Fout << "L2Width\t";
	Fout << "L1Height\t";
	Fout << "L1Width\t";
	Fout << "L3Height\t";
	Fout << "L3Width\t";
	Fout << "Byte24:0F\t";
	Fout << "Palette\t";
	Fout << "Byte26\t";
	Fout << "Byte27:1F\t";
	Fout << "Byte27:E0\t";
	Fout << "Byte28\t";
	Fout << "Byte29\t";
	Fout << "Byte30\t";
	Fout << "Byte31\t";
	Fout << "Byte32\n";
	
	for(i = 0; i < 416; i++){
		Fout << i << "\t";
		Fout << (ushort) RomData[0x2D9100 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9101 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9102 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9103 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9104 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9105 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9106 + (i * 33)] << "\t";
		Fout << (RomData[0x2D9107 + (i * 33)] & 0x7F) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D9107 + (i * 33)]) & 0x3F80) >> 7) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D9108 + (i * 33)]) & 0x1FC0) >> 6) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D9109 + (i * 33)]) & 0x0FE0) >> 5) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D910A + (i * 33)]) & 0x03F0) >> 4) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D910B + (i * 33)]) & 0x01FC) >> 2) << "\t";
		Fout << ((RomData[0x2D910C + (i * 33)] & 0xFE) >> 1) << "\t";
		Fout << (*((ushort*) &RomData[0x2D910D + (i * 33)]) & 0x03FF) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D910E + (i * 33)]) & 0x0FFC) >> 2) << "\t";
		Fout << ((*((ushort*) &RomData[0x2D910F + (i * 33)]) & 0x3FF0) >> 4) << "\t";
		Fout << (ushort) RomData[0x2D9111 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9112 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9113 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9114 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9115 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9116 + (i * 33)] << "\t";
		Fout << (1 << (4 + (RomData[0x2D9117 + (i * 33)] & 0x03))) << "\t";
		Fout << (1 << (4 + ((RomData[0x2D9117 + (i * 33)] & 0x0C) >> 2))) << "\t";
		Fout << (1 << (4 + ((RomData[0x2D9117 + (i * 33)] & 0x30) >> 4))) << "\t";
		Fout << (1 << (4 + ((RomData[0x2D9117 + (i * 33)] & 0xC0) >> 6))) << "\t";
		Fout << (1 << (4 + ((RomData[0x2D9118 + (i * 33)] & 0x30) >> 4))) << "\t";
		Fout << (1 << (4 + ((RomData[0x2D9118 + (i * 33)] & 0xC0) >> 6))) << "\t";
		Fout << (RomData[0x2D9118 + (i * 33)] & 0x0F) << "\t";
		Fout << (ushort) RomData[0x2D9119 + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D911A + (i * 33)] << "\t";
		Fout << (RomData[0x2D911B + (i * 33)] & 0x1F) << "\t";
		Fout << ((RomData[0x2D911B + (i * 33)] & 0xE0) >> 5) << "\t";
		Fout << (ushort) RomData[0x2D911C + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D911D + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D911E + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D911F + (i * 33)] << "\t";
		Fout << (ushort) RomData[0x2D9120 + (i * 33)] << "\n";
	}
	Fout.close();
}



void CFF3Doc::WriteLocStrings(){
	int i;
	ofstream Fout;
	uint nStrLen;
	string sLocName;

	Fout.open("locstrings.txt");

	Fout << "Index\tName\n";
	for(i = 0; i < 73; i++){
		Fout << i << "\t";
		uint nAddr = ((RomData[0x268601 + (i << 1)] << 8) | RomData[0x268600 + (i << 1)]) + 0x0EF300;

		nStrLen = 0;
		sLocName = "";
		while(RomData[nAddr++] != 0x00){
			nStrLen++;
		}
		TranslateString2(&RomData[nAddr - (nStrLen + 1)], &sLocName, nStrLen);
		Fout << sLocName << "\n";
	}
	Fout.close();
}



void CFF3Doc::MakeNoCompression(byte* RawData, uint nLength, char* sFilename){
	ofstream Fout;
	Fout.open(sFilename, ios_base::binary);
	byte nPacketHeader = 0xFF;
	for(uint i = 0; i < nLength; i++){
		if(!(i & 0x07)){
			Fout.write((const char*) &nPacketHeader, 1);
		}
		Fout.write((const char*) &RawData[i], 1);
	}
	Fout.close();
}



void CFF3Doc::WriteBattleBG(){
	byte* BatBGGraphics[75];
	byte BatBGExt[3][4096];
	byte* BatPackPtr[3];
	memset(BatBGGraphics, 0, 75 << 2);
	uint nGraphicAddr = 0;
	OBBitmap BatBGPack;
	BatBGPack.SetDim(128, 64);
	BatBGPack.SetPaletteGrey(16);
	BatBGPack.nBitPlanes = 4;

	OBBitmap BatBG;
	BatBG.SetDim(256, 152);
	BatBG.SetPaletteGrey(16);
	BatBG.nBitPlanes = 4;

	byte nPacket;
	byte nShift;
	byte bSkip;
	int i;
	char filename[64];
	byte RawData[65536];
	for(i = 0; i < 56; i++){
		for(byte j = 0; j < 3; j++){
			if(RomData[0x270200 + (i * 6) + j] == 0xFF){
				memset(BatBGExt[j], 0, 4096);
				nGraphicAddr = (*((uint*) &RomData[0x271850 + (nPacket & 0x7F) * 3]) & 0x00FFFFFF) - 0xBFFE00 + 0x1000;
				memcpy(BatBGExt[j], &RomData[nGraphicAddr], 4096);
				BatPackPtr[j] = BatBGExt[j];
				BatBGPack.ProcessData(BatBGExt[j]);
				sprintf(filename, "battlebackgrounds\\packet%02XExt%d.bmp", nPacket & 0x7F, j);
				BatBGPack.WriteOutBitmap(filename);
			}
			else{
				nPacket = RomData[0x270200 + (i * 6) + j];
				if(!BatBGGraphics[nPacket & 0x7F]){
					BatBGGraphics[nPacket & 0x7F] = new byte[4096];
					memset(BatBGGraphics[nPacket & 0x7F], 0, 4096);
					nGraphicAddr = (*((uint*) &RomData[0x271850 + (nPacket & 0x7F) * 3]) & 0x00FFFFFF) - 0xBFFE00;
					if(nPacket & 0x80){
						memcpy(BatBGGraphics[nPacket & 0x7F], &RomData[nGraphicAddr], 4096);
					}
					else{
						DecompressData(&RomData[nGraphicAddr], BatBGGraphics[nPacket & 0x7F], 4096);
					}
					BatBGPack.ProcessData(BatBGGraphics[nPacket & 0x7F]);
					sprintf(filename, "battlebackgrounds\\packet%02X.bmp", nPacket & 0x7F);
					BatBGPack.WriteOutBitmap(filename);
				}
				BatPackPtr[j] = BatBGGraphics[nPacket & 0x7F];
			}
		}
		DecompressData(&RomData[*((ushort*) &RomData[0x271A48 + (RomData[0x270200 + (i * 6) + 3] << 1)]) + 0x270200], RawData);
		BatBG.SetDefaultColor(0);
		BatBG.SetPalette((ushort*) &RomData[0x270350 + 3 * ((RomData[0x270200 + (i * 6) + 5] & 0x7F) << 5)], 48);
		for(byte y = 0; y < 19; y++){
			for(byte x = 0; x < 32; x++){
				bSkip = 0;
				switch(*((ushort*) &RawData[(y << 6) + (x << 1)]) & 0x0180){
				case 0x0000:
					nPacket = 0;
					break;
				case 0x0080:
					nPacket = 1;
					break;
				case 0x0100:
				case 0x0180:
					nPacket = 2;
					break;
				default:
					bSkip = 1;
					break;
				}
				switch(RawData[(y << 6) + (x << 1) + 1] & 0x0C){
				case 0x04:
					nShift = 0x00;
					break;
				case 0x08:
					nShift = 0x01;
					break;
				case 0x0C:
					nShift = 0x02;
					break;
				default:
					nShift++;
					break;
				}
				if(!bSkip){
					BatBG.Put8x8Chunk(x << 3, y << 3, &BatPackPtr[nPacket][(RawData[(y << 6) + (x << 1)] & 0x7F) << 5], nShift, (RawData[(y << 6) + (x << 1) + 1] & 0xC0) >> 6);
				}
			}
		}
		sprintf(filename, "battlebackgrounds\\BatBG%02X.bmp", i);
		BatBG.WriteOutBitmap(filename);
	}
}